# The software in this package is distributed under the GNU General
# Public License version 2 (with a special exception described below).
#
# A copy of GNU General Public License (GPL) is included in this distribution,
# in the file COPYING.GPL.
#
# As a special exception, if other files instantiate templates or use macros
# or inline functions from this file, or you compile this file and link it
# with other works to produce a work based on this file, this file
# does not by itself cause the resulting work to be covered
# by the GNU General Public License.
#
# However the source code for this file must still be made available
# in accordance with section (3) of the GNU General Public License.
#
# This exception does not invalidate any other reasons why a work based
# on this file might be covered by the GNU General Public License.

""" test_helper_unittest.py: unit tests for test_helpers

Tests classes and functions in test_helpers

Should be able run from python2 and python3!

For help see :py:mod:`unittest`

.. codeauthor:: Intra2net
"""

from __future__ import print_function

import unittest
from datetime import timedelta, datetime as dt
from file_helpers import FilesystemFillState, get_filesystem_fill_states
import warnings
from time import sleep
import threading

import test_helpers


class TestHelperTester(unittest.TestCase):

    def assertNoWarn(self, func, *args, **kwargs):
        """ check that calling function raises no warning

        use warnings.catch_warnings instead of self.assertWarn
        for compatibility with python 2
        """

        result = None
        with warnings.catch_warnings(record=True) as warn_catcher:
            result = func(*args, **kwargs)

            self.assertEqual(len(warn_catcher), 0)

        return result


    def assertWarn(self, warn_type, n_warns, message, func, *args, **kwargs):
        """ check that calling function raises no warning

        use warnings.catch_warnings instead of self.assertWarn
        for compatibility with python 2
        """

        with warnings.catch_warnings(record=True) as warn_catcher:
            func(*args, **kwargs)

            if n_warns is None:
                self.assertTrue(len(warn_catcher) > 0)
            else:
                self.assertEqual(len(warn_catcher), n_warns)

            if warn_type is not None:
                self.assertTrue(issubclass(warn_catcher[-1].category,
                                           warn_type))
            if message is not None:
                self.assertEqual(warn_catcher[-1].message, message)
            #print(warn_catcher[-1].message)


    def test_zero_estimator(self):
        """ test function :py:func:`test_helpers.calc_time_until_zero` """

        # need to allow some tolerance since calc_time_until_zero uses now()
        TOLERANCE = 0.1

        start_time = dt.now()
        time1 = start_time - timedelta(seconds=20)    # 20s ago
        time2 = start_time - timedelta(seconds=10)    # 10s ago

        # test constant disc usage
        self.assertEqual(test_helpers.calc_time_until_zero((time1, 1),
                                                           (time2, 1)), None)

        # test gaining disc space (1 byte per second, empty 30s ago)
        estim = test_helpers.calc_time_until_zero((time1, 10), (time2, 20))
        self.assertTrue(estim < 0)
        expect = ((time1-timedelta(seconds=10)) - dt.now()).total_seconds()
        self.assertLess(abs(estim - expect), TOLERANCE)

        # test slowly losing disc space (1 byte per second, empty now)
        estim = test_helpers.calc_time_until_zero((time1, 20), (time2, 10))
        expect = (start_time - dt.now()).total_seconds()
        self.assertLess(abs(estim - expect), TOLERANCE)


    def test_decision_function(self):
        """ tests function default_disc_full_decision_function

        :py:func:`test_helpers.default_disc_full_decision_function`
        """

        f = test_helpers.default_disc_full_decision_function

        # create dummy state
        GIGABYTE = 2**30
        state = FilesystemFillState()
        state.name = 'dummy'
        state.size = 10 * GIGABYTE
        state.used = 9 * GIGABYTE
        state.available = 1 * GIGABYTE
        state.capacity = 90
        state.mount_point = '/not/mounted'

        HOUR = 60*60
        estim_empty_newest = None    # currently not used
        estim_empty_mean = None      # currently not used
        estim_empty_smallest = None  # simulate no proper estimate

        # check no warning if no time estimate (and enough space)
        self.assertNoWarn(f, state, estim_empty_newest, estim_empty_smallest,
                          estim_empty_mean)

        # check no warning if enough time (and enough space)
        estim_empty_smallest = HOUR      # enough time
        self.assertNoWarn(f, state, estim_empty_newest, estim_empty_smallest,
                          estim_empty_mean)

        # check warn if time getting low
        estim_empty_smallest = (test_helpers.DEFAULT_TIME_TO_FULL_WARN +
                                test_helpers.DEFAULT_TIME_TO_FULL_ERR) / 2.
        self.assertWarn(test_helpers.DiscFullPreventionWarning, 1, None,
                        f, state, estim_empty_newest, estim_empty_smallest,
                        estim_empty_mean)

        # check err if time too low
        estim_empty_smallest = test_helpers.DEFAULT_TIME_TO_FULL_ERR / 2.
        self.assertRaises(KeyboardInterrupt,
                          f, state, estim_empty_newest, estim_empty_smallest,
                          estim_empty_mean)
        # show the error message:
        #f(state, estim_empty_newest, estim_empty_smallest, estim_empty_mean)

        # check warning if disc space is getting low
        estim_empty_smallest = HOUR
        state.available = int((test_helpers.DEFAULT_DISC_SIZE_WARN +
                               test_helpers.DEFAULT_DISC_SIZE_ERR ) / 2.)
        state.used = state.size - state.available
        state.capacity = int(100. * state.used / state.size)
        self.assertWarn(test_helpers.DiscFullPreventionWarning, 1, None,
                        f, state, estim_empty_newest, estim_empty_smallest,
                        estim_empty_mean)

        # check error if disc space is too low
        state.available = int(test_helpers.DEFAULT_DISC_SIZE_ERR / 2.)
        state.used = state.size - state.available
        state.capacity = int(100. * state.used / state.size)
        self.assertRaises(KeyboardInterrupt,
                          f, state, estim_empty_newest, estim_empty_smallest,
                          estim_empty_mean)

        # check other kill levels
        self.assertRaises(test_helpers.DiscFullPreventionError,
                          f, state, estim_empty_newest, estim_empty_smallest,
                          estim_empty_mean,
                          kill_level=test_helpers.KILL_EXCEPTION)

        self.assertRaises(SystemExit,
                          f, state, estim_empty_newest, estim_empty_smallest,
                          estim_empty_mean,
                          kill_level=test_helpers.KILL_SYS_EXIT)

        # this even kills unittest:
        #f(state, estim_empty_newest, estim_empty_smallest, estim_empty_mean,
        #  kill_level=test_helpers.KILL_BRUTALLY_ONLY_LAST_RESORT)
        #self.fail('should have exited!')

    def test_thread(self):
        """ test the DiscFillCheckerThread """

        #threading.setprofile(self.threading_profile_func)
        # somehow damages df output ... need to check!

        # since we do not want to fill up the disc, we incrementally decrease
        # the tolerance so even our hopefully half-empty filesystems will
        # trigger warnings / errors
        min_free = None
        min_name = None
        for fill_state in get_filesystem_fill_states():
            if fill_state.size == 0:
                continue
            if (min_free is None) or (fill_state.available < min_free):
                min_free = fill_state.available
                min_name = fill_state.name
        print('min free: {0} with {1}B'.format(min_name, min_free))

        if min_free is None:
            self.fail('Found no filesystem with size>0 --> cannot test')

        # set space tolerance such that no warn/err should show
        size_warn = min_free/4;
        size_err = min_free/2;
        def decision_function(*args):
            test_helpers.default_disc_full_decision_function(
                *args, size_warn=size_warn, size_err=size_err)
        self._internal_test_thread(decision_function)

        # set space tolerance such that show warn but no err
        size_warn = min_free*2;
        size_err = min_free/2;
        def decision_function(*args):
            test_helpers.default_disc_full_decision_function(
                *args, size_warn=size_warn, size_err=size_err)
        self.assertWarn(test_helpers.DiscFullPreventionWarning, None, None,
                        self._internal_test_thread, decision_function)

        # set space tolerance such that err should be raised
        size_warn = min_free*4;
        size_err = min_free*2;
        def decision_function(*args):
            test_helpers.default_disc_full_decision_function(
                *args, size_warn=size_warn, size_err=size_err)
        self.assertRaises(KeyboardInterrupt,
                          self._internal_test_thread, decision_function)

    def _internal_test_thread(self, decision_function):

        sleep_time_checker = 0.1
        sleep_time_tester = 1
        time_tolerance = 0.1
        main_thread_name = 'MainThread'
        checker_thread_name = 'discFillChck'

        curr_threads = [thread.name for thread in threading.enumerate()]
        self.assertEqual(curr_threads, [main_thread_name])
        start_time = dt.now()
        with test_helpers.disc_fill_checked(test_helpers.METHOD_THREAD,
                                            sleep_time_checker,
                                            decision_function) as checker:
            curr_threads = [thread.name for thread in threading.enumerate()]
            self.nonsense_function(sleep_time_tester)
        time_taken = (dt.now() - start_time).total_seconds()
        self.assertEqual(curr_threads, [main_thread_name, checker_thread_name])

        # stop checker thread
        checker.stop()
        checker.join()
        curr_threads = [thread.name for thread in threading.enumerate()]
        self.assertEqual(curr_threads, [main_thread_name])

        self.assertTrue(time_taken >= sleep_time_tester)
        self.assertTrue(time_taken < sleep_time_tester + time_tolerance)


    def nonsense_function(self, sleep_time):
        """ function to run while checking disc fill state; just sleeps

        to make this test quicker could try to use profiling to count calls to
        disc checker function and stop this after 2 or so
        """
        sleep(sleep_time)

    def threading_profile_func(self, *args, **kwargs):
        """ todo: set this as profiling function for checker thread, count
        calls to disc checker function, trigger event or so """
        print('profiling with args {0} and kwargs {1}'
              .format([arg for arg in args if len(args) < 20], kwargs))

if __name__ == '__main__':
    unittest.main()
